/*
 * netdump_client.c
 * FreeBSD kernel module supporting netdump network dumps.
 * netdump_server must be running to accept client dumps.
 *
 * Copyright (c) 2000 Darrell Anderson <anderson@cs.duke.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/sysctl.h>
#include <sys/lock.h>
#include <sys/reboot.h>
#include <sys/eventhandler.h>
#include <sys/module.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_var.h>
#include <netinet/in_var.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>
#include <netinet/ip_fw.h>
#include <vm/vm.h>
#include <vm/vm_object.h>
#ifdef KLD_MODULE
#undef KLD_MODULE /* yuck. need PQ_FREE */
#include <vm/vm_page.h>
#define KLD_MODULE
#else
#include <vm/vm_page.h>
#endif
#include <machine/md_var.h>
#include <machine/clock.h>
#include <machine/pcb.h>

#include "netdump.h"

extern struct pcb dumppcb; /* cheat.  dumppcb is a static! */

/* ---------------------------------------------------------------- */
/*
 * private globals.  don't touch.
 */
static struct in_addr nd_server = {INADDR_BROADCAST}; /* server address */
static caddr_t nd_kva = NULL;                /* copy data here before xmit */
static eventhandler_tag nd_tag = NULL;       /* record of our shutdown event */
static u_int32_t nd_seqno = 1, nd_ackno = 0; /* sequence and ack numbers */
static ip_fw_chk_t *old_ip_fw_chk_ptr;       /* previous ip_fw_chk_ptr */

/* ---------------------------------------------------------------- */
/*
 * sysctl pokables.
 */

static int nd_enable = 1;  /* perform a network dump */
static int nd_partial = 1; /* partial dump -- skip uninteresting pages */

SYSCTL_NODE(_net, OID_AUTO, dump, CTLFLAG_RW, 0, "netdump");
SYSCTL_INT(_net_dump, OID_AUTO, enable, CTLFLAG_RW, &nd_enable, 0, "enable network dump");
SYSCTL_INT(_net_dump, OID_AUTO, partial, CTLFLAG_RW, &nd_partial, 0, "enable network dump");

/* ---------------------------------------------------------------- */

/*
 * [netdump_output]
 *
 * unreliable transmission of an mbuf chain to the specified ip or
 * broadcast address.
 *
 * Parameters:
 *      target  ip or broadcast address (network byte order)
 *      m       mbuf chain
 *
 * Returns:
 *      int     see errno.h, 0 for success
 */
static int
netdump_output(struct in_addr target, struct mbuf *m)
{
	struct udpiphdr *ui;
	struct ip *ip;
	struct route ro;
	struct sockaddr_in *src = NULL, *dst;
	struct in_ifaddr *ia;
	int error;

	M_PREPEND(m, sizeof(struct udpiphdr), M_DONTWAIT);
	if (m == 0) {
		return ENOBUFS;
	}

	/*
	 * compute the route early so we can grab the interface address
	 * as our packet return address (need before checksumming).
	 */
	dst = (struct sockaddr_in *)&ro.ro_dst;
	dst->sin_family = AF_INET;
	dst->sin_len = sizeof(struct sockaddr_in);
	dst->sin_addr = target;
	ro.ro_rt = (struct rtentry *)0;
	rtalloc_ign(&ro, RTF_CLONING);
	for (ia = in_ifaddrhead.tqh_first; ia; ia = ia->ia_link.tqe_next) {
		if (ia->ia_ifp == ro.ro_rt->rt_ifp) {
			src = IA_SIN(ia);
			break;
		}
	}
	if (src == NULL) {
		return EHOSTUNREACH;
	}

	/*
	 * fill in the udp header (including checksum), then the ip dregs.
	 */
	ui = mtod(m, struct udpiphdr *);
	bzero(ui->ui_x1, sizeof(ui->ui_x1));
        ui->ui_pr = IPPROTO_UDP;
        ui->ui_len = htons(m->m_pkthdr.len - sizeof(struct ip));
        ui->ui_src = src->sin_addr;
        ui->ui_dst = dst->sin_addr;
        ui->ui_sport = ui->ui_dport = htons(NETDUMP_PORT);
        ui->ui_ulen = ui->ui_len;
	ui->ui_sum = 0;
	if ((ui->ui_sum = in_cksum(m, m->m_pkthdr.len)) == 0)
                ui->ui_sum = 0xffff;
	ip = mtod(m, struct ip *);
	ip->ip_len = m->m_pkthdr.len;
	ip->ip_off = ip->ip_tos = 0;
	ip->ip_ttl = 32;

	error = ip_output(m, NULL, &ro, IP_ALLOWBROADCAST, 0);
	rtfree(ro.ro_rt); /* route stays cached */
	return error;
}

/* ---------------------------------------------------------------- */
/*
 * this section provides reliable message delivery (in the absence
 * of resource shortages).
 */

int fxp_intr __P((void *));
void tulip_intr_normal __P((void *));

/*
 * [netdump_network_poll]
 *
 * after trapping, interrupts go the way of the dodo.  we need to receive
 * incoming packets (acks) to ensure reliable delivery, so poll for
 * interrupts.  there must be a better way to do this.  XXX
 * (enable_intr() does NOT do the trick..)
 *
 * Parameters:
 *      void
 *
 * Returns:
 *      void
 */
static void
netdump_network_poll(void)
{
        struct ifnet *ifn;
        struct ifaddr *ifa;
        struct in_addr addr;
	int s;

	s = splnet();	
        if ((ifn = TAILQ_FIRST(&ifnet)) != NULL) do {
                if ((ifn->if_flags & IFF_UP) == 0) {
                        continue;
		}
                if ((ifa = TAILQ_FIRST(&ifn->if_addrhead)) != NULL) do {
                        if (ifa->ifa_addr->sa_family != AF_INET) {
                                continue;
			}
                        addr = ((struct sockaddr_in *)ifa->ifa_addr)->sin_addr;
                        if (addr.s_addr == INADDR_LOOPBACK) {
                                continue;
			}

			if (strcmp(ifn->if_name, "fxp") == 0) {
				fxp_intr(ifn->if_softc);
			} else if (strcmp(ifn->if_name, "de") == 0) {
				tulip_intr_normal(ifn->if_softc);
			}

                } while ((ifa = TAILQ_NEXT(ifa, ifa_link)) != NULL);
        } while ((ifn = TAILQ_NEXT(ifn, if_link)) != NULL);
	splx(s);
}


/*
 * [netdump_mbuf_nop]
 *
 * netdump wraps external mbufs around address ranges.  unlike most sane
 * counterparts, netdump uses a stop-and-wait approach to flow control and
 * retransmission, so the ack obviates the need for mbuf reference
 * counting.  we still need to tell other mbuf handlers not to do anything
 * special with our mbufs, so specify this nop handler.
 *
 * Parameters:
 *      addr    mbuf data address (ignored)
 *      len     mbuf data length (ignored)
 *
 * Returns:
 *      void
 */
static void
netdump_mbuf_nop(caddr_t addr, u_int len)
{
	;
}

/*
 * [netdump_send]
 *
 * construct and reliably send a netdump packet.  may fail from a resource
 * shortage or extreme number of unacknowledged retransmissions.  wait for
 * an acknowledgement before returning.
 *
 * Parameters:
 *      target  ip or broadcast address (network byte order)
 *      type    netdump packet type (HERALD, FINISHED, or VMCORE)
 *      offset  vmcore data offset (bytes)
 *      data    vmcore data
 *      datalen vmcore data size (bytes)
 *
 * Returns:
 *      int see errno.h, 0 for success
 */
static int
netdump_send(struct in_addr target, u_int32_t type, u_int32_t offset, 
	     unsigned char *data, u_int32_t datalen)
{
	struct netdump_msg_hdr *nd_msg_hdr;
	struct mbuf *m, *m2;
	int retries = 0, polls, error;

 retransmit:

	/*
	 * get and fill a header mbuf, then chain data as an extended mbuf.
	 */
	if ((m = m_gethdr(M_DONTWAIT, MT_DATA)) == NULL) {
		return ENOBUFS;
	}
	m->m_pkthdr.len = m->m_len = sizeof(struct netdump_msg_hdr);
	MH_ALIGN(m, sizeof(struct netdump_msg_hdr)); /* leave room for udpip */
	nd_msg_hdr = mtod(m, struct netdump_msg_hdr *);
	nd_msg_hdr->seqno = htonl(nd_seqno);
	nd_msg_hdr->type = htonl(type);
	nd_msg_hdr->offset = htonl(offset);
	nd_msg_hdr->len = htonl(datalen);

	if (datalen) {
		if ((m2 = m_get(M_DONTWAIT, MT_DATA)) == NULL) {
			m_freem(m);
			return ENOBUFS;
		}
		m2->m_flags |= M_EXT;
		m2->m_data = m2->m_ext.ext_buf = data;
		m2->m_len = datalen;
		m2->m_ext.ext_ref = m2->m_ext.ext_free = netdump_mbuf_nop;
		m->m_next = m2;
		m->m_pkthdr.len += m2->m_len;
	}

	if ((error = netdump_output(target, m)) != 0) {
		printf("netdump_send returned %d\n", error);
		return error;
	}

	/*
	 * wait for an ack.  a window would help bandwidth.
	 */
	polls = 0;
	while (nd_ackno < nd_seqno) {		
		if (polls++ > 1000) {
			if (retries++ > 10) {
				return ETIMEDOUT; /* 10 s, no ack */
			}
			printf(". ");
			goto retransmit; /* 1 s, no ack */
		}
		DELAY(1000); /* 1 ms */
		/*
		 * this is not always necessary, but does no harm.
		 */
		netdump_network_poll();
	}
	nd_seqno++;
	return 0;
}

/* ---------------------------------------------------------------- */

/*
 * [netdump_ipfw_hook]
 *
 * hook to examine incoming and outgoing ip packets.  identify incoming
 * netdump ack packets and adjust the ack sequence number.  also note the
 * server address on the first response (we broadcasted to find it).
 *
 * Parameters:
 *	ipp	Pointer to packet header (struct ip **)
 *	hlen	Packet header length
 *	ifp	Outgoing interface, or NULL if packet is incoming
 *	*cookie Skip up to the first rule past this rule number;
 *	*m	The packet; we set to NULL when/if we nuke it.
 *	*flow_id pointer to the last matching rule (in/out)
 *	*next_hop socket we are forwarding to (in/out).
 *
 * Return value:
 *	0	The packet is to be accepted and routed normally OR
 *      	the packet was denied/rejected and has been dropped;
 *		in the latter case, *m is equal to NULL upon return.
 *      port    Divert the packet to port.  
 */
static int
netdump_ipfw_hook (struct ip_fw_args *args)
{
	struct mbuf *m = args->m;
	struct udpiphdr *udp;
	struct ip *ip;
	struct netdump_ack *nd_ack;
	struct mbuf **mp = &(args->m);

	ip = mtod(m, struct ip *); 
	if (m->m_pkthdr.rcvif == NULL) {
		goto pass; /* outgoing packet -- not interested */
	}
	if (m->m_pkthdr.len <
	    sizeof(struct udpiphdr) + sizeof(struct netdump_ack)) {
		goto pass; /* too small -- not interested */
	}
	if (ip->ip_p != IPPROTO_UDP) {
		goto pass; /* not udp -- not interested */
	}

	if (m->m_len <
	    sizeof(struct udpiphdr) + sizeof(struct netdump_ack)) {
		int s = splimp();
		*mp = m = m_pullup(m, sizeof(struct udpiphdr) + 
				   sizeof(struct netdump_ack));
		ip = mtod(m, struct ip *);
		splx(s);
	}
	udp = mtod(m, struct udpiphdr *);

	if (ntohs(udp->ui_u.uh_dport) != NETDUMP_ACKPORT) {
		/* printf("sport = %d, dport = %d, len %d\n",
		    ntohs(udp->ui_u.uh_sport),
		    ntohs(udp->ui_u.uh_dport),
		    ntohs(udp->ui_u.uh_ulen));*/

		goto pass; /* not ours -- not interested */
	}
	/*
	 * packet is meant for us.  extract the ack sequence number.
	 * if it's the first ack, extract the server address as well.
	 */
	nd_ack = (struct netdump_ack *)
		(mtod(m, caddr_t) + sizeof(struct udpiphdr));
	if (nd_server.s_addr == INADDR_BROADCAST) {
		nd_server = ip->ip_src;
	}
	nd_ackno = ntohl(nd_ack->seqno);	
	
	m_freem(m);
	*mp = NULL; /* we consumed the packet.  stop further processing. */
	return 0;

 pass:
#if 0 /* we're crashing.  don't play well with strangers. */
	if (old_ip_fw_chk_ptr) {
		return (*old_ip_fw_chk_ptr)(ipp, hlen, ifp, cookie, mp, rule, addrp);
	}
#endif
	return 0;
}

/* ---------------------------------------------------------------- */

extern int debugger_on_panic;
/*
 * [netdump_trigger]
 *
 * called from kern_shutdown during "boot" (invoked on panic).  perform a
 * network dump, and if successful cancel the normal disk dump.
 *
 * Parameters:
 *      arg     unused
 *      howto   boot flags (only dump if RB_DUMP set)
 *
 * Returns:
 *      void
 */
static void
netdump_trigger(void *arg, int howto)
{
	int sendcount, skipcount, ticks_before, ticks_after, secs;
	off_t offset, memsize;
	vm_page_t m;
	char *va;
	static int active = 0;
	
	if ((howto&(RB_HALT|RB_DUMP))!=RB_DUMP || !nd_enable || active++) {
		return;
	}

	/*
	 * netdump is invoked as a shutdown handler instead of as
	 * a real dumpdev dump routine.  (networks are shut down
	 * before dumpsys() gets called.)  make sure the dump context
	 * is set so a debugger can find the stack trace.
	 */
	savectx(&dumppcb);
#ifdef __i386__
	dumppcb.pcb_cr3 = rcr3();
#endif

	/* 
	 * install our incoming packet hook and broadcast the dump intent
	 * (nd_server initially INADDR_BROADCAST).  when the first
	 * netdump_send returns, nd_server contains the server addr.
	 * deinstall after the netdump completes.
	 */
	old_ip_fw_chk_ptr = ip_fw_chk_ptr;
	ip_fw_chk_ptr = netdump_ipfw_hook;

#ifdef __i386__
	/*
	 * panic() leaves interrupts disabled.  re-enable so we can
	 * receive packets.  (harmless if interrupts were already on.)
	 * not necessary under FreeBSD/alpha.
	 */
	enable_intr();
#endif
	/*
	 * for good measure, drop to spl0 so software interrupts
	 * (e.g. ip_input) can run.
	 */
	spl0();

	printf("\n-----------------------------------\n");
	printf("netdump in progress. searching for server.. ");
	if (netdump_send(nd_server, NETDUMP_HERALD, 0, NULL, 0) != 0) {
		goto abort;
	}
	printf("dumping to %s\n", inet_ntoa(nd_server));
	printf("-----------------------------------\n");

	/*
	 * dump memory.
	 */
	ticks_before = ticks;
	sendcount = skipcount = 0;
	memsize = Maxmem * PAGE_SIZE;
	for (offset=0 ; offset < memsize ; offset += PAGE_SIZE) {

		if (offset % (1024*1024) == 0) {
#ifdef HW_WDOG
			if (wdog_tickler)
				(*wdog_tickler)();
#endif
			printf("%ld ", (long)(memsize-offset) / (1024*1024));
		}

		/*
		 * a normal crash dumps all of physical memory, but on a
		 * large memory machine the dump and subsequent savecore
		 * dominates the reboot cycle.  optionally only write (and
		 * savecore) interesting pages.  uninteresting pages
		 * incluce (1.) holes in the physical address space, (2.)
		 * free pages, and (3.) vnode backed pages.  preserve the
		 * physical offsets of pages we do write so that an
		 * unmodified savecore or other utility can still use the
		 * dump.
		 */
		if (nd_partial &&
		    ((long long)(memsize-offset) / (1024ULL*1024ULL)) > 10 &&
		    ((is_physical_memory(offset) == 0) ||
		     (m = PHYS_TO_VM_PAGE(trunc_page(offset))) == NULL ||
		     (m->queue - m->pc == PQ_FREE) ||
		     (m->object && m->object->type == OBJT_VNODE))) {
			skipcount++;
			continue;
		}

		/*
		 * temporarily map the physical offset, copy to our private
		 * space, and unmap before sending.  need to copy because
		 * page may change between computing the checksum and
		 * sending the data.
		 */
		va = pmap_kenter_temporary(trunc_page(offset), 0);
		bcopy(va, nd_kva, PAGE_SIZE);
	
		if (netdump_send(nd_server, NETDUMP_VMCORE, offset, nd_kva,
				 PAGE_SIZE) != 0) {
			goto abort;
		}
		sendcount++;
	}
	if (netdump_send(nd_server, NETDUMP_FINISHED, 0, NULL, 0) != 0) {
		goto abort;
	}
	ticks_after = ticks;

	printf("\nnetdump finished, wrote %d pages, skipped %d.\n", 
	       sendcount, skipcount);

	if ((secs = (ticks_after - ticks_before) / hz) != 0) {
		printf("[wrote %d MB in %d seconds]\n",
		       (sendcount*PAGE_SIZE)/(1024*1024), secs);
	}

	if (dumpdev != NODEV) {
		printf("cancelling normal dump\n");
		dumpdev = NODEV;
	}
	ip_fw_chk_ptr = old_ip_fw_chk_ptr;
	return;
 abort:
	ip_fw_chk_ptr = old_ip_fw_chk_ptr;
	printf("\nnetdump failed, proceeding to normal dump\n");
}

/*------------------------------------------------*/
/*
 * module specific handling
 */

static int
netdump_modevent(module_t mod, int type, void *unused) 
{
	switch (type) {
	case MOD_LOAD:
		MALLOC(nd_kva, caddr_t, PAGE_SIZE, M_DEVBUF, M_WAITOK);
		if (nd_kva == NULL) {
			printf("failed to allocate netdump_kva\n");
			return ENOSPC;
		}
		/* PRI_FIRST happens before the networks are disabled */
		nd_tag = EVENTHANDLER_REGISTER(shutdown_pre_sync, 
					       netdump_trigger, NULL, 
					       SHUTDOWN_PRI_FIRST);
		printf("netdump installed\n");
		break;
	case MOD_UNLOAD:
		if (nd_kva) {
			FREE(nd_kva, M_DEVBUF);
			nd_kva = NULL;
		}
		if (nd_tag) {
			EVENTHANDLER_DEREGISTER(shutdown_pre_sync, nd_tag);
			nd_tag = NULL;
		}
		printf("netdump deinstalled\n");
		break;
	default:
		break;
	}
	return 0;
}
static moduledata_t netdump_mod = {"netdump", netdump_modevent, 0};
DECLARE_MODULE(netdump, netdump_mod, SI_SUB_DRIVERS, SI_ORDER_ANY);
