/*
 * netdump_server.c
 * user level server that accepts netdump network dumps.
 *
 * Copyright (c) 2000 Darrell Anderson <anderson@cs.duke.edu>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <sys/errno.h>

#include "netdump.h"

static char *prefix;
static char dumpname[128];

/*
 * [ip2name]
 *
 * translate a host ip address to a hostname.  exit verbosely on error.
 *
 * Parameters:
 *      ip      ip address (network byte order)
 *
 * Returns:
 *      char *  hostname (null-terminated ascii string)
 */
static char *
ip2name(struct in_addr ip)
{
	static char name[128];
	struct hostent *he;
	char *dot;
	
	if ((he = gethostbyaddr((caddr_t)&ip, sizeof(ip), AF_INET)) == NULL) {
		fprintf(stderr, "can't resolve %d.%d.%d.%d\n",
			(ip.s_addr>>24)&0xff, (ip.s_addr>>16)&0xff,
			(ip.s_addr>>8)&0xff, (ip.s_addr)&0xff);
		exit(1);
	}
	strncpy(name, he->h_name, sizeof(name));
	if ((dot = strchr(name, '.')) != NULL) {
		*dot = '\0';
	}
	return name;
}

/*
 * [name2fd]
 *
 * given a hostname, generate a file descriptor for its vmcore file.
 *
 * Parameters:
 *      name    hostname (null-terminated ascii string)
 *
 * Returns:
 *      int     file descriptor for vmcore file
 */
static int
name2fd(char *name)
{
	int fd;
#if 0
	struct stat sb;
	int i;

	for (i=0 ; i<256 ; i++) {
		snprintf(dumpname, sizeof(dumpname), "%s/vmcore.%s.%d", 
			 prefix, name, i);
		if (stat(dumpname, &sb) == 0) {
			continue; /* file existed, proceed. */
		}
		if (errno != ENOENT) {
			perror("stat"); /* unexpected error, die. */
			exit(1);
		}
		if ((fd = open(dumpname, O_WRONLY|O_CREAT, DEFFILEMODE)) == -1) {
			fprintf(stderr, "open failed for \"%s\"\n", dumpname);
			perror("open");
			exit(1);
		}
		return fd;
	}
	printf("too many %s/%s.vmcore.X files!\n", prefix, name);
	exit(1);
	return -1; /* unreachable -- stop compiler warnings. */
#else
	snprintf(dumpname, sizeof(dumpname), "%s/vmcore.%s", 
		 prefix, name);
	if ((fd = open(dumpname, O_TRUNC|O_WRONLY|O_CREAT, DEFFILEMODE)) == -1) {
		fprintf(stderr, "open failed for \"%s\"\n", dumpname);
		perror("open");
		exit(1);
	}
	return fd;
#endif
}

/*
 * [sockaddr_init]
 *
 * fill in a sockaddr given its component fields.
 *
 * Parameters:
 *      saddr   sockaddr
 *      ipaddr  ip address (network byte order)
 *      port    port (network byte order)
 *
 * Returns:
 *      void
 */
static void
sockaddr_init(struct sockaddr_in *saddr, u_int32_t ipaddr, short port)
{
	bzero(saddr, sizeof(struct sockaddr_in));
	saddr->sin_family = AF_INET;
	saddr->sin_len = sizeof(struct sockaddr_in);
	saddr->sin_port = port;
	saddr->sin_addr.s_addr = ipaddr;
}

/*
 * [serverloop]
 *
 * main program loop.  open a connectionless socket listening on the
 * netdump port and loop receiving packets from anyone.  process each
 * packet independently.
 *
 * Parameters:
 *      void
 *
 * Returns:
 *      void
 */
static void
serverloop(void)
{
	int sock, fd, cnt, rcvflag = 0/*MSG_WAITALL*/;
	struct sockaddr_in local, frm;
	struct in_addr cached_ip;
	u_int32_t frmlen;
	struct netdump_msg *nd_msg;
	struct netdump_ack *nd_ack;
	char *name;
	
	nd_msg = (struct netdump_msg *)malloc(sizeof(struct netdump_msg));
	nd_ack = (struct netdump_ack *)malloc(sizeof(struct netdump_ack));
	if (nd_msg == NULL || nd_ack == NULL) {
		perror("malloc");
		exit(1);
	}
	if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1) {
		perror("socket");
		exit(1);
	}
	sockaddr_init(&local, INADDR_ANY, htons(NETDUMP_PORT));
	if (bind(sock, (struct sockaddr *)&local, local.sin_len) == -1) {
		perror("bind");
		exit(1);
	}
	printf("netdump_server listening on port %d, writing vmcores to %s\n",
	       NETDUMP_PORT, prefix);

	while (1) {

		/*
		 * get a packet
		 */
		sockaddr_init(&frm, INADDR_ANY, 0);
		frmlen = sizeof(struct sockaddr_in);
		if ((cnt = recvfrom(sock, nd_msg, sizeof(*nd_msg), rcvflag, 
				    (struct sockaddr *)&frm, &frmlen)) == -1) {
			perror("recvfrom");
			exit(1);
		}
		if (cnt < sizeof(struct netdump_msg_hdr) +
		    ntohl(nd_msg->hdr.len)) {
			printf("runt: read %d, expected %d\n", cnt,
			       (int)sizeof(struct netdump_msg_hdr) +
			       (int)ntohl(nd_msg->hdr.len));
		}

		/*
		 * always start fresh for a herald packet.
		 */
		if (ntohl(nd_msg->hdr.type) == NETDUMP_HERALD && fd) {
			close (fd);
			fd = 0;
			cached_ip.s_addr = 0;
		}

		/*
		 * get a vmcore file descriptor (if not already cached)
		 */
		if (frm.sin_addr.s_addr != cached_ip.s_addr) {
			if (fd) {
				close (fd);
			}
			name = ip2name(frm.sin_addr);
			fd = name2fd(name);
			cached_ip.s_addr = frm.sin_addr.s_addr;
		}

		/*
		 * process the packet
		 */
		switch(ntohl(nd_msg->hdr.type)) {
		case NETDUMP_HERALD:
			printf("starting dump from %s to %s\n", name, dumpname);
			if (ftruncate(fd, 0) == -1) {
				perror("ftruncate");
				exit(1);
			}
			break;
		case NETDUMP_FINISHED:
			printf("\nfinished dump from %s\n", name);
			close(fd); /* close fd so unlink will work */
			cached_ip.s_addr = 0; /* invalidate cache */
			break;
		case NETDUMP_VMCORE:
			if (ntohl(nd_msg->hdr.seqno) % 100 == 0) {
				printf(".");
				fflush(stdout);
			}
			if (pwrite(fd, nd_msg->data, ntohl(nd_msg->hdr.len),
				   ntohl(nd_msg->hdr.offset)) == -1) {
				perror("pwrite");
				exit(1);
			}
			break;
		default:
			fprintf(stderr, "nd_msg bad type %d\n", 
				(int)ntohl(nd_msg->hdr.type));
			exit(1);
		}
		
		/*
		 * send an ack. frm has return address, but needs ack port.
		 */
		frm.sin_port = htons(NETDUMP_ACKPORT);
		nd_ack->seqno = nd_msg->hdr.seqno; /* already network order */
		if (sendto(sock, nd_ack, sizeof(struct netdump_ack), 0,
			   (struct sockaddr *)&frm, frm.sin_len) == -1) {
			perror("sendto");
			/*exit(1);*/
		}

	}
}

int
main(int argc, char *argv[])
{
	struct stat sb;
	if (argc != 2) {
		fprintf(stderr, "usage: %s vmcore_path\n", argv[0]);
		exit(1);
	}
	prefix = argv[1];
	if (stat(prefix, &sb) == -1) {
		perror("stat");
		fprintf(stderr, "invalid vmcore_path \"%s\"\n", prefix);
		exit(1);
	}
	serverloop(); /* never returns */
	return 0;
}
